@@ -66,13 +66,14 @@ from six.moves import queue
 from six.moves import urllib
 
 from . import errors, __version__
+from ._compat import ntob
 from .workers import threadpool
 from .makefile import MakeFile
 
 
-__all__ = ['HTTPRequest', 'HTTPConnection', 'HTTPServer',
+__all__ = ('HTTPRequest', 'HTTPConnection', 'HTTPServer',
            'SizeCheckWrapper', 'KnownLengthRFile', 'ChunkedRFile',
-           'Gateway', 'get_ssl_adapter_class']
+           'Gateway', 'get_ssl_adapter_class')
 
 
 if 'win' in sys.platform and hasattr(socket, 'AF_INET6'):
@@ -82,26 +83,6 @@ if 'win' in sys.platform and hasattr(socket, 'AF_INET6'):
         socket.IPV6_V6ONLY = 27
 
 
-if six.PY3:
-    def ntob(n, encoding='ISO-8859-1'):
-        """Return the given native string as a byte string in the given encoding."""
-        # In Python 3, the native string type is unicode
-        return n.encode(encoding)
-
-    def bton(b, encoding='ISO-8859-1'):
-        return b.decode(encoding)
-else:
-    def ntob(n, encoding='ISO-8859-1'):
-        """Return the given native string as a byte string in the given encoding."""
-        # In Python 2, the native string type is bytes. Assume it's already
-        # in the given encoding, which for ISO-8859-1 is almost always what
-        # was intended.
-        return n
-
-    def bton(b, encoding='ISO-8859-1'):
-        return b
-
-
 LF = b'\n'
 CRLF = b'\r\n'
 TAB = b'\t'
@@ -131,59 +112,89 @@ if not hasattr(logging, 'statistics'):
     logging.statistics = {}
 
 
-def read_headers(rfile, hdict=None):
-    """Read headers from the given stream into the given header dict.
+class HeaderReader(object):
+    """Object for reading headers from an HTTP request.
 
-    If hdict is None, a new header dict is created. Returns the populated
-    header dict.
-
-    Headers which are repeated are folded together using a comma if their
-    specification so dictates.
-
-    This function raises ValueError when the read bytes violate the HTTP spec.
-    You should probably return "400 Bad Request" if this happens.
+    Interface and default implementation.
     """
-    if hdict is None:
-        hdict = {}
 
-    while True:
-        line = rfile.readline()
-        if not line:
-            # No more data--illegal end of headers
-            raise ValueError('Illegal end of headers.')
+    def __call__(self, rfile, hdict=None):
+        """
+        Read headers from the given stream into the given header dict.
 
-        if line == CRLF:
-            # Normal end of headers
-            break
-        if not line.endswith(CRLF):
-            raise ValueError('HTTP requires CRLF terminators')
+        If hdict is None, a new header dict is created. Returns the populated
+        header dict.
 
-        if line[0] in (SPACE, TAB):
-            # It's a continuation line.
-            v = line.strip()
-        else:
-            try:
-                k, v = line.split(COLON, 1)
-            except ValueError:
-                raise ValueError('Illegal header line.')
-            # TODO: what about TE and WWW-Authenticate?
-            k = k.strip().title()
-            v = v.strip()
-            hname = k
+        Headers which are repeated are folded together using a comma if their
+        specification so dictates.
 
-        if k in comma_separated_headers:
-            existing = hdict.get(hname)
-            if existing:
-                v = b', '.join((existing, v))
-        hdict[hname] = v
+        This function raises ValueError when the read bytes violate the HTTP spec.
+        You should probably return "400 Bad Request" if this happens.
+        """
+        if hdict is None:
+            hdict = {}
 
-    return hdict
+        while True:
+            line = rfile.readline()
+            if not line:
+                # No more data--illegal end of headers
+                raise ValueError('Illegal end of headers.')
+
+            if line == CRLF:
+                # Normal end of headers
+                break
+            if not line.endswith(CRLF):
+                raise ValueError('HTTP requires CRLF terminators')
+
+            if line[0] in (SPACE, TAB):
+                # It's a continuation line.
+                v = line.strip()
+            else:
+                try:
+                    k, v = line.split(COLON, 1)
+                except ValueError:
+                    raise ValueError('Illegal header line.')
+                v = v.strip()
+                k = self._transform_key(k)
+                hname = k
+
+            if not self._allow_header(k):
+                continue
+
+            if k in comma_separated_headers:
+                existing = hdict.get(hname)
+                if existing:
+                    v = b', '.join((existing, v))
+            hdict[hname] = v
+
+        return hdict
+
+    def _allow_header(self, key_name):
+        return True
+
+    def _transform_key(self, key_name):
+        # TODO: what about TE and WWW-Authenticate?
+        return key_name.strip().title()
+
+
+class DropUnderscoreHeaderReader(HeaderReader):
+    """Custom HeaderReader to exclude any headers with underscores in them."""
+
+    def _allow_header(self, key_name):
+        orig = super(DropUnderscoreHeaderReader, self)._allow_header(key_name)
+        return orig and '_' not in key_name
 
 
 class SizeCheckWrapper(object):
     """Wraps a file-like object, raising MaxSizeExceeded if too large."""
 
     def __init__(self, rfile, maxlen):
+        """Initialize SizeCheckWrapper instance.
+
+        Args:
+            rfile (file): file of a limited size
+            maxlen (int): maximum length of the file being read
+        """
         self.rfile = rfile
         self.maxlen = maxlen
         self.bytes_read = 0
@@ -193,12 +204,28 @@ class SizeCheckWrapper(object):
             raise errors.MaxSizeExceeded()
 
     def read(self, size=None):
+        """Read a chunk from rfile buffer and return it.
+
+        Args:
+            size (int): amount of data to read
+
+        Returns:
+            bytes: Chunk from rfile, limited by size if specified.
+        """
         data = self.rfile.read(size)
         self.bytes_read += len(data)
         self._check_length()
         return data
 
     def readline(self, size=None):
+        """Read a single line from rfile buffer and return it.
+
+        Args:
+            size (int): minimum amount of data to read
+
+        Returns:
+            bytes: One line from rfile.
+        """
         if size is not None:
             data = self.rfile.readline(size)
             self.bytes_read += len(data)
@@ -218,69 +245,14 @@ class SizeCheckWrapper(object):
                 return EMPTY.join(res)
 
     def readlines(self, sizehint=0):
-        # Shamelessly stolen from StringIO
-        total = 0
-        lines = []
-        line = self.readline()
-        while line:
-            lines.append(line)
-            total += len(line)
-            if 0 < sizehint <= total:
-                break
-            line = self.readline()
-        return lines
+        """Read all lines from rfile buffer and return them.
 
-    def close(self):
-        self.rfile.close()
+        Args:
+            sizehint (int): hint of minimum amount of data to read
 
-    def __iter__(self):
-        return self
-
-    def __next__(self):
-        data = next(self.rfile)
-        self.bytes_read += len(data)
-        self._check_length()
-        return data
-
-    def next(self):
-        data = self.rfile.next()
-        self.bytes_read += len(data)
-        self._check_length()
-        return data
-
-
-class KnownLengthRFile(object):
-    """Wraps a file-like object, returning an empty string when exhausted."""
-
-    def __init__(self, rfile, content_length):
-        self.rfile = rfile
-        self.remaining = content_length
-
-    def read(self, size=None):
-        if self.remaining == 0:
-            return b''
-        if size is None:
-            size = self.remaining
-        else:
-            size = min(size, self.remaining)
-
-        data = self.rfile.read(size)
-        self.remaining -= len(data)
-        return data
-
-    def readline(self, size=None):
-        if self.remaining == 0:
-            return b''
-        if size is None:
-            size = self.remaining
-        else:
-            size = min(size, self.remaining)
-
-        data = self.rfile.readline(size)
-        self.remaining -= len(data)
-        return data
-
-    def readlines(self, sizehint=0):
+        Returns:
+            list[bytes]: Lines of bytes read from rfile.
+        """
         # Shamelessly stolen from StringIO
         total = 0
         lines = []
@@ -294,12 +266,112 @@ class KnownLengthRFile(object):
         return lines
 
     def close(self):
+        """Release resources allocated for rfile."""
         self.rfile.close()
 
     def __iter__(self):
+        """Return file iterator."""
         return self
 
     def __next__(self):
+        """Generate next file chunk."""
+        data = next(self.rfile)
+        self.bytes_read += len(data)
+        self._check_length()
+        return data
+
+    def next(self):
+        """Generate next file chunk."""
+        data = self.rfile.next()
+        self.bytes_read += len(data)
+        self._check_length()
+        return data
+
+
+class KnownLengthRFile(object):
+    """Wraps a file-like object, returning an empty string when exhausted."""
+
+    def __init__(self, rfile, content_length):
+        """Initialize KnownLengthRFile instance.
+
+        Args:
+            rfile (file): file of a known size
+            content_length (int): length of the file being read
+        """
+        self.rfile = rfile
+        self.remaining = content_length
+
+    def read(self, size=None):
+        """Read a chunk from rfile buffer and return it.
+
+        Args:
+            size (int): amount of data to read
+
+        Returns:
+            bytes: Chunk from rfile, limited by size if specified.
+        """
+        if self.remaining == 0:
+            return b''
+        if size is None:
+            size = self.remaining
+        else:
+            size = min(size, self.remaining)
+
+        data = self.rfile.read(size)
+        self.remaining -= len(data)
+        return data
+
+    def readline(self, size=None):
+        """Read a single line from rfile buffer and return it.
+
+        Args:
+            size (int): minimum amount of data to read
+
+        Returns:
+            bytes: One line from rfile.
+        """
+        if self.remaining == 0:
+            return b''
+        if size is None:
+            size = self.remaining
+        else:
+            size = min(size, self.remaining)
+
+        data = self.rfile.readline(size)
+        self.remaining -= len(data)
+        return data
+
+    def readlines(self, sizehint=0):
+        """Read all lines from rfile buffer and return them.
+
+        Args:
+            sizehint (int): hint of minimum amount of data to read
+
+        Returns:
+            list[bytes]: Lines of bytes read from rfile.
+        """
+        # Shamelessly stolen from StringIO
+        total = 0
+        lines = []
+        line = self.readline(sizehint)
+        while line:
+            lines.append(line)
+            total += len(line)
+            if 0 < sizehint <= total:
+                break
+            line = self.readline(sizehint)
+        return lines
+
+    def close(self):
+        """Release resources allocated for rfile."""
+        self.rfile.close()
+
+    def __iter__(self):
+        """Return file iterator."""
+        return self
+
+    def __next__(self):
+        """Generate next file chunk."""
         data = next(self.rfile)
         self.remaining -= len(data)
         return data
@@ -314,6 +386,13 @@ class ChunkedRFile(object):
     """
 
     def __init__(self, rfile, maxlen, bufsize=8192):
+        """Initialize ChunkedRFile instance.
+
+        Args:
+            rfile (file): file encoded with the 'chunked' transfer encoding
+            maxlen (int): maximum length of the file being read
+            bufsize (int): size of the buffer used to read the file
+        """
         self.rfile = rfile
         self.maxlen = maxlen
         self.bytes_read = 0
@@ -359,7 +438,19 @@ class ChunkedRFile(object):
                 'got ' + repr(crlf) + ')')
 
     def read(self, size=None):
+        """Read a chunk from rfile buffer and return it.
+
+        Args:
+            size (int): amount of data to read
+
+        Returns:
+            bytes: Chunk from rfile, limited by size if specified.
+        """
         data = EMPTY
+
+        if size == 0:
+            return data
+
         while True:
             if size and len(data) >= size:
                 return data
@@ -376,9 +467,22 @@ class ChunkedRFile(object):
                 self.buffer = self.buffer[remaining:]
             else:
                 data += self.buffer
+                self.buffer = EMPTY
 
     def readline(self, size=None):
+        """Read a single line from rfile buffer and return it.
+
+        Args:
+            size (int): minimum amount of data to read
+
+        Returns:
+            bytes: One line from rfile.
+        """
         data = EMPTY
+
+        if size == 0:
+            return data
+
         while True:
             if size and len(data) >= size:
                 return data
@@ -402,11 +506,20 @@ class ChunkedRFile(object):
             else:
                 if newline_pos == -1:
                     data += self.buffer
+                    self.buffer = EMPTY
                 else:
                     data += self.buffer[:newline_pos]
                     self.buffer = self.buffer[newline_pos:]
 
     def readlines(self, sizehint=0):
+        """Read all lines from rfile buffer and return them.
+
+        Args:
+            sizehint (int): hint of minimum amount of data to read
+
+        Returns:
+            list[bytes]: Lines of bytes read from rfile.
+        """
         # Shamelessly stolen from StringIO
         total = 0
         lines = []
@@ -420,6 +533,11 @@ class ChunkedRFile(object):
         return lines
 
     def read_trailer_lines(self):
+        """Read HTTP headers and yield them.
+
+        Returns:
+            Generator: yields CRLF separated lines.
+        """
         if not self.closed:
             raise ValueError(
                 'Cannot read trailers until the request body has been read.')
@@ -443,6 +561,7 @@ class ChunkedRFile(object):
             yield line
 
     def close(self):
+        """Release resources allocated for rfile."""
         self.rfile.close()
 
 
@@ -479,7 +598,18 @@ class HTTPRequest(object):
 
     This value is set automatically inside send_headers."""
 
+    header_reader = HeaderReader()
+    """
+    A HeaderReader instance or compatible reader.
+    """
+
     def __init__(self, server, conn):
+        """Initialize HTTP request container instance.
+
+        Args:
+            server (HTTPServer): web server object receiving this request
+            conn (HTTPConnection): HTTP connection object for this request
+        """
         self.server = server
         self.conn = conn
 
@@ -530,6 +660,11 @@ class HTTPRequest(object):
         self.ready = True
 
     def read_request_line(self):
+        """Read and parse first line of the HTTP request.
+
+        Returns:
+            bool: True if the request line is valid or False if it's malformed.
+        """
         # HTTP/1.1 connections are persistent by default. If a client
         # requests a page, then idles (leaves the connection open),
         # then rfile.readline() will raise socket.error("timed out").
@@ -600,8 +735,7 @@ class HTTPRequest(object):
                 unquote_to_bytes(x)
                 for x in quoted_slash.split(path)
             ]
-        except ValueError:
-            ex = sys.exc_info()[1]
+        except ValueError as ex:
             self.simple_response('400 Bad Request', ex.args[0])
             return False
         path = b'%2F'.join(atoms)
@@ -638,9 +772,8 @@ class HTTPRequest(object):
         """Read self.rfile into self.inheaders. Return success."""
         # then all the http headers
         try:
-            read_headers(self.rfile, self.inheaders)
-        except ValueError:
-            ex = sys.exc_info()[1]
+            self.header_reader(self.rfile, self.inheaders)
+        except ValueError as ex:
             self.simple_response('400 Bad Request', ex.args[0])
             return False
 
@@ -707,9 +840,8 @@ class HTTPRequest(object):
             msg += b' 100 Continue\r\n\r\n'
             try:
                 self.conn.wfile.write(msg)
-            except socket.error:
-                x = sys.exc_info()[1]
-                if x.args[0] not in errors.socket_errors_to_ignore:
+            except socket.error as ex:
+                if ex.args[0] not in errors.socket_errors_to_ignore:
                     raise
         return True
 
@@ -808,9 +940,8 @@ class HTTPRequest(object):
 
         try:
             self.conn.wfile.write(EMPTY.join(buf))
-        except socket.error:
-            x = sys.exc_info()[1]
-            if x.args[0] not in errors.esocket_errors_to_ignore:
+        except socket.error as ex:
+            if ex.args[0] not in errors.socket_errors_to_ignore:
                 raise
 
     def write(self, chunk):
@@ -896,12 +1027,7 @@ class HTTPRequest(object):
 
 
 class HTTPConnection(object):
-    """An HTTP connection (active socket).
-
-    server: the Server object which received this connection.
-    socket: the raw socket object (usually TCP) for this connection.
-    makefile: a fileobject class for reading from the socket.
-    """
+    """An HTTP connection (active socket)."""
 
     remote_addr = None
     remote_port = None
@@ -911,6 +1037,13 @@ class HTTPConnection(object):
     RequestHandlerClass = HTTPRequest
 
     def __init__(self, server, sock, makefile=MakeFile):
+        """Initialize HTTPConnection instance.
+
+        Args:
+            server (HTTPServer): web server object receiving this request
+            socket (socket._socketobject): the raw socket object (usually TCP) for this connection
+            makefile (file): a fileobject class for reading from the socket
+        """
         self.server = server
         self.socket = sock
         self.rfile = makefile(sock, 'rb', self.rbufsize)
@@ -942,9 +1075,8 @@ class HTTPConnection(object):
                 req.respond()
                 if req.close_connection:
                     return
-        except socket.error:
-            e = sys.exc_info()[1]
-            errnum = e.args[0]
+        except socket.error as ex:
+            errnum = ex.args[0]
             # sadly SSL sockets return a different (longer) time out string
             if (
                 errnum == 'timed out' or
@@ -964,7 +1096,7 @@ class HTTPConnection(object):
                             # Close the connection.
                             return
                         except errors.NoSSLError:
-                            self._handle_no_ssl()
+                            self._handle_no_ssl(req)
             elif errnum not in errors.socket_errors_to_ignore:
                 self.server.error_log('socket.error %s' % repr(errnum),
                                       level=logging.WARNING, traceback=True)
@@ -975,7 +1107,7 @@ class HTTPConnection(object):
                         # Close the connection.
                         return
                     except errors.NoSSLError:
-                        self._handle_no_ssl()
+                        self._handle_no_ssl(req)
             return
         except (KeyboardInterrupt, SystemExit):
             raise
@@ -984,9 +1116,8 @@ class HTTPConnection(object):
             return
         except errors.NoSSLError:
             self._handle_no_ssl(req)
-        except Exception:
-            e = sys.exc_info()[1]
-            self.server.error_log(repr(e), level=logging.ERROR, traceback=True)
+        except Exception as ex:
+            self.server.error_log(repr(ex), level=logging.ERROR, traceback=True)
             if req and not req.sent_headers:
                 try:
                     req.simple_response('500 Internal Server Error')
@@ -1085,7 +1216,7 @@ class HTTPServer(object):
     """The maximum number of worker threads to create (default -1 = no limit)."""
 
     server_name = None
-    """The name of the server; defaults to socket.gethostname()."""
+    """The name of the server; defaults to ``self.version``."""
 
     protocol = 'HTTP/1.1'
     """The version string to write in the Status-Line of all HTTP responses.
@@ -1132,19 +1263,35 @@ class HTTPServer(object):
     You must have the corresponding SSL driver library installed.
     """
 
-    def __init__(self, bind_addr, gateway, minthreads=10, maxthreads=-1,
-                 server_name=None):
+    def __init__(self, bind_addr, gateway, minthreads=10, maxthreads=-1, server_name=None):
+        """Initialize HTTPServer instance.
+
+        Args:
+            bind_addr (tuple): network interface to listen to
+            gateway (Gateway): gateway for processing HTTP requests
+            minthreads (int): minimum number of threads for HTTP thread pool
+            maxthreads (int): maximum number of threads for HTTP thread pool
+            server_name (str): web server name to be advertised via Server HTTP header
+        """
         self.bind_addr = bind_addr
         self.gateway = gateway
 
         self.requests = threadpool.ThreadPool(self, min=minthreads or 1, max=maxthreads)
 
         if not server_name:
-            server_name = socket.gethostname()
+            server_name = self.version
+            """
+            Default server name. It is used to fall back to ``socket.gethostname()``.
+            Ref: cherrypy/cheroot#33. If the previous logic needed the following should restore it:
+
+            >>> from cheroot._compat import import ntou, tonative
+            >>> ntou(tonative(ntou(tonative(hn, 'utf-8'), 'utf-8').encode('idna')))
+            """
         self.server_name = server_name
         self.clear_stats()
 
     def clear_stats(self):
+        """Reset server stat counters.."""
         self._start_time = None
         self._run_time = 0
         self.stats = {
@@ -1174,15 +1321,17 @@ class HTTPServer(object):
                  for w in s['Worker Threads'].values()], 0),
             'Worker Threads': {},
         }
-        logging.statistics['CherryPy HTTPServer %d' % id(self)] = self.stats
+        logging.statistics['Cheroot HTTPServer %d' % id(self)] = self.stats
 
     def runtime(self):
+        """Return server uptime."""
         if self._start_time is None:
             return self._run_time
         else:
             return self._run_time + (time.time() - self._start_time)
 
     def __str__(self):
+        """Render Server instance representing bind address."""
         return '%s.%s(%r)' % (self.__module__, self.__class__.__name__,
                               self.bind_addr)
 
@@ -1329,6 +1478,13 @@ class HTTPServer(object):
                     raise self.interrupt
 
     def error_log(self, msg='', level=20, traceback=False):
+        """Write error message to log.
+
+        Args:
+            msg (str): error message
+            level (int): logging level
+            traceback (bool): add traceback to output or not
+        """
         # Override this in subclasses as desired
         sys.stderr.write(msg + '\n')
         sys.stderr.flush()
@@ -1395,9 +1551,8 @@ class HTTPServer(object):
                     wfile = mf(sock_to_make, 'wb', io.DEFAULT_BUFFER_SIZE)
                     try:
                         wfile.write(''.join(buf).encode('ISO-8859-1'))
-                    except socket.error:
-                        x = sys.exc_info()[1]
-                        if x.args[0] not in errors.socket_errors_to_ignore:
+                    except socket.error as ex:
+                        if ex.args[0] not in errors.socket_errors_to_ignore:
                             raise
                     return
                 if not s:
@@ -1436,11 +1591,10 @@ class HTTPServer(object):
             # notice keyboard interrupts on Win32, which don't interrupt
             # accept() by default
             return
-        except socket.error:
-            x = sys.exc_info()[1]
+        except socket.error as ex:
             if self.stats['Enabled']:
                 self.stats['Socket Errors'] += 1
-            if x.args[0] in errors.socket_error_eintr:
+            if ex.args[0] in errors.socket_error_eintr:
                 # I *think* this is right. EINTR should occur when a signal
                 # is received during the accept() call; all docs say retry
                 # the call, and I *think* I'm reading it right that Python
@@ -1448,11 +1602,11 @@ class HTTPServer(object):
                 # elsewhere. See
                 # https://github.com/cherrypy/cherrypy/issues/707.
                 return
-            if x.args[0] in errors.socket_errors_nonblocking:
+            if ex.args[0] in errors.socket_errors_nonblocking:
                 # Just try again. See
                 # https://github.com/cherrypy/cherrypy/issues/479.
                 return
-            if x.args[0] in errors.socket_errors_to_ignore:
+            if ex.args[0] in errors.socket_errors_to_ignore:
                 # Our socket was closed.
                 # See https://github.com/cherrypy/cherrypy/issues/686.
                 return
@@ -1482,9 +1636,8 @@ class HTTPServer(object):
                 # Touch our own socket to make accept() return immediately.
                 try:
                     host, port = sock.getsockname()[:2]
-                except socket.error:
-                    x = sys.exc_info()[1]
-                    if x.args[0] not in errors.socket_errors_to_ignore:
+                except socket.error as ex:
+                    if ex.args[0] not in errors.socket_errors_to_ignore:
                         # Changed to use error code and not message
                         # See
                         # https://github.com/cherrypy/cherrypy/issues/860.
@@ -1520,6 +1673,11 @@ class Gateway(object):
     """A base class to interface HTTPServer with other systems, such as WSGI."""
 
     def __init__(self, req):
+        """Initialize Gateway instance with request.
+
+        Args:
+            req (HTTPRequest): current HTTP request
+        """
         self.req = req
 
     def respond(self):
